@@ -754,7 +754,7 @@ void handle_quota_options(ext2_filsys fs)
 
 void parse_quota_opts(const char *opts)
 {
-	char	*buf, *token, *next, *p, *arg;
+	char	*buf, *token, *next, *p;
 	int	len;
 
 	len = strlen(opts);
@@ -1925,6 +1925,7 @@ retry_open:
 		ext2fs_free(fs);
 		exit(1);
 	}
+	fs->default_bitmap_type = EXT2FS_BMAP64_RBTREE;
 
 	if (I_flag && !io_ptr_orig) {
 		/*
